@@ -19,49 +19,78 @@ import (
 	"io"
 	"math/rand"
 	"net"
+	"os"
 	"regexp"
 	"strings"
 
-	"github.com/miekg/dns"
 	"github.com/pkg/errors"
 	log "github.com/sirupsen/logrus"
+	"github.com/zmap/dns"
 	"github.com/zmap/zcrypto/tls"
 	"github.com/zmap/zcrypto/x509"
 	"github.com/zmap/zgrab2/lib/http"
 	"github.com/zmap/zgrab2/lib/output"
 
-	"github.com/zmap/zdns/src/internal/util"
+	"github.com/zmap/zdns/v2/src/internal/util"
 )
 
 var ErrorContextExpired = errors.New("context expired")
 
-// GetDNSServers returns a list of IPv4, IPv6 DNS servers from a file, or an error if one occurs
 func GetDNSServers(path string) (ipv4, ipv6 []string, err error) {
-	c, err := dns.ClientConfigFromFile(path)
+	file, err := os.Open(path)
 	if err != nil {
-		return []string{}, []string{}, fmt.Errorf("error reading DNS config file (%s): %w", path, err)
+		return nil, nil, fmt.Errorf("error opening DNS config file (%s): %w", path, err)
+	}
+	defer func(file *os.File) {
+		err := file.Close()
+		if err != nil {
+			log.Errorf("error closing DNS config file (%s): %s", path, err)
+		}
+	}(file)
+	return getDNSServersFromReader(file)
+}
+
+// getDNSServersFromReader returns a list of IPv4, IPv6 DNS servers from an io.Reader, or an error if one occurs
+func getDNSServersFromReader(resolvReader io.Reader) (ipv4, ipv6 []string, err error) {
+	c, err := dns.ClientConfigFromReader(resolvReader)
+	if err != nil {
+		return []string{}, []string{}, fmt.Errorf("error parsing DNS config file: %v", err)
 	}
 	servers := make([]string, 0, len(c.Servers))
-	for _, s := range c.Servers {
-		if s[0:1] != "[" && strings.Contains(s, ":") {
-			s = "[" + s + "]"
-		}
-		full := strings.Join([]string{s, c.Port}, ":")
-		servers = append(servers, full)
-	}
 	ipv4 = make([]string, 0, len(servers))
 	ipv6 = make([]string, 0, len(servers))
-	for _, s := range servers {
-		ip, _, err := util.SplitHostPort(s)
-		if err != nil {
-			return []string{}, []string{}, fmt.Errorf("could not parse IP address (%s) from file: %w", s, err)
+	for _, s := range c.Servers {
+		// We don't support specifying link-local IPv6 addresses with %interface or domain names with #domain
+		// See https://man7.org/linux/man-pages/man1/resolvectl.1.html
+		if strings.Contains(s, "%") {
+			return []string{}, []string{}, fmt.Errorf("could not parse IP address (%s) from config. We do not support specifying link-local IPv6 addresses or per-interface nameservers", s)
 		}
-		if ip.To4() != nil {
-			ipv4 = append(ipv4, s)
-		} else if util.IsIPv6(&ip) {
-			ipv6 = append(ipv6, s)
+		if strings.Contains(s, "#") {
+			return []string{}, []string{}, fmt.Errorf("could not parse IP address (%s) from config. We do not support specifying domain names for nameservers", s)
+		}
+		// We need to check if there is a port specified, and add the default if not
+		ipStr, _, err := net.SplitHostPort(s)
+		if err == nil {
+			// port specified, determine IP type
+			ip := net.ParseIP(ipStr)
+			if ip != nil && ip.To4() != nil {
+				ipv4 = append(ipv4, s)
+			} else if ip != nil {
+				ipv6 = append(ipv6, s)
+			}
+			continue
+		}
+		// no port specified, check if s is an IP
+		ip := net.ParseIP(s)
+		if ip == nil {
+			return []string{}, []string{}, fmt.Errorf("could not parse IP address (%s) from config", s)
+		} else if ip.To4() != nil {
+			// IPv4, use default port
+			ipv4 = append(ipv4, strings.Join([]string{s, c.Port}, ":"))
 		} else {
-			return []string{}, []string{}, fmt.Errorf("could not parse IP address (%s) from file: %s", s, path)
+			// IPv6, use default port
+			s = "[" + s + "]"
+			ipv6 = append(ipv6, strings.Join([]string{s, c.Port}, ":"))
 		}
 	}
 	return ipv4, ipv6, nil
@@ -80,6 +109,9 @@ func (lc LookupClient) DoDstServersLookup(ctx context.Context, r *Resolver, q Qu
 }
 
 func (r *Resolver) doDstServersLookup(ctx context.Context, q Question, nameServers []NameServer, isIterative bool) (*SingleQueryResult, Trace, Status, error) {
+	if util.HasCtxExpired(ctx) {
+		return nil, nil, StatusTimeout, ErrorContextExpired
+	}
 	var err error
 	// nameserver is required
 	if len(nameServers) == 0 {
@@ -91,7 +123,7 @@ func (r *Resolver) doDstServersLookup(ctx context.Context, q Question, nameServe
 		qname, err = dns.ReverseAddr(q.Name)
 		// might be an actual DNS name instead of an IP address
 		// if that looks likely, use it as is
-		if err != nil && !util.IsStringValidDomainName(q.Name) {
+		if err != nil && !util.IsStringValidDomainName(qname) {
 			return nil, nil, StatusIllegalInput, err
 			// q.Name is a valid name, we can continue
 		} else {
@@ -274,8 +306,8 @@ func isLookupComplete(originalName string, candidateSet map[string][]Answer, cNa
 
 // LookupAllNameserversExternal will query all nameServers with the given question and return the results
 // If nameServers is empty, it will use the externalNameServers from the resolver
-func (r *Resolver) LookupAllNameserversExternal(q *Question, nameServers []NameServer) ([]SingleQueryResult, Trace, Status, error) {
-	ctx, cancel := context.WithTimeout(context.Background(), r.timeout)
+func (r *Resolver) LookupAllNameserversExternal(ctx context.Context, q *Question, nameServers []NameServer) ([]SingleQueryResult, Trace, Status, error) {
+	ctx, cancel := context.WithTimeout(ctx, r.timeout)
 	defer cancel()
 	retv := make([]SingleQueryResult, 0)
 	var trace Trace
@@ -366,9 +398,9 @@ func (r *Resolver) filterNameServersForUniqueNames(nameServers []NameServer) []N
 //
 // Additionally, we'll query each layer for NS records, and once we have the set of authoritative nameservers, we'll query with
 // the original question type. This helps find sibling nameservers that aren't listed with the TLD.
-func (r *Resolver) LookupAllNameserversIterative(q *Question, rootNameServers []NameServer) (*AllNameServersResult, Trace, Status, error) {
+func (r *Resolver) LookupAllNameserversIterative(ctx context.Context, q *Question, rootNameServers []NameServer) (*AllNameServersResult, Trace, Status, error) {
 	perNameServerRetriesLimit := 2
-	ctx, cancel := context.WithTimeout(context.Background(), r.timeout)
+	ctx, cancel := context.WithTimeout(ctx, r.timeout)
 	defer cancel()
 	retv := AllNameServersResult{
 		LayeredResponses: make(map[string][]ExtendedResult),
@@ -755,6 +787,9 @@ func (r *Resolver) cyclingLookup(ctx context.Context, qWithMeta *QuestionWithMet
 		} else if *qWithMeta.RetriesRemaining == 0 {
 			r.verboseLog(depth+1, "Cycling lookup failed - out of retries. Name: ", qWithMeta.Q.Name, ", Layer: ", layer, ", Nameserver: ", nameServer)
 			return result, isCached, status, trace, errors.New("cycling lookup failed - out of retries")
+		} else if util.HasCtxExpired(ctx) {
+			r.verboseLog(depth+1, "Cycling lookup failed - context expired. Name: ", qWithMeta.Q.Name, ", Layer: ", layer, ", Nameserver: ", nameServer)
+			return result, isCached, status, trace, errors.New("cycling lookup failed - context expired")
 		} else if !isStatusRetryable(status) {
 			r.verboseLog(depth+1, "Cycling lookup failed - unretryable status:", status, "Name: ", qWithMeta.Q.Name, ", Layer: ", layer, ", Nameserver: ", nameServer)
 			return result, isCached, status, trace, err
@@ -790,6 +825,9 @@ func getRandomNonQueriedNameServer(nameServers []NameServer, queriedNameServers
 // cacheBasedOnNameServer is whether to consider a cache hit based on DNS question and nameserver, or just question
 // cacheNonAuthoritative is whether to cache non-authoritative answers, usually used for lookups using an external resolver
 func (r *Resolver) cachedLookup(ctx context.Context, q Question, nameServer *NameServer, layer string, depth int, requestIteration, cacheBasedOnNameServer, cacheNonAuthoritative bool, trace Trace) (*SingleQueryResult, IsCached, Status, Trace, error) {
+	if util.HasCtxExpired(ctx) {
+		return &SingleQueryResult{}, false, StatusTimeout, trace, ErrorContextExpired
+	}
 	// check for circular queries. This may be problematic if NS has circular references and we're trying to perform a DNSSEC validation
 	if _, ok := r.pendingQueries[q]; ok {
 		return &SingleQueryResult{}, false, StatusCircular, trace, errors.New("circular query detected")
@@ -942,6 +980,9 @@ func (r *Resolver) cachedLookup(ctx context.Context, q Question, nameServer *Nam
 }
 
 func doDoTLookup(ctx context.Context, connInfo *ConnectionInfo, q Question, nameServer *NameServer, rootCAs *x509.CertPool, shouldVerifyServerCert, recursive bool, ednsOptions []dns.EDNS0, dnssec bool, checkingDisabled bool) (*SingleQueryResult, *dns.Msg, Status, error) {
+	if util.HasCtxExpired(ctx) {
+		return nil, nil, StatusTimeout, errors.New("context expired")
+	}
 	m := new(dns.Msg)
 	m.SetQuestion(dotName(q.Name), q.Type)
 	m.Question[0].Qclass = q.Class
@@ -1128,7 +1169,7 @@ func wireLookupTCP(ctx context.Context, connInfo *ConnectionInfo, q Question, na
 		if err != nil && err.Error() == "EOF" {
 			// EOF error means the connection was closed, we'll remove the connection (it'll be recreated on the next iteration)
 			// and try again
-			err = connInfo.tcpConn.Conn.Close()
+			err = connInfo.tcpConn.Close()
 			if err != nil {
 				log.Errorf("error closing TCP connection: %v", err)
 			}
